#=======================================================================================================
# PACKAGE NAME: VerifyingUserChoices
#    FUNCTIONS:	 - InfoForUser (Give back information to the user),
#				 - AlignmentCheck (Verify if the alignment is done for clustering steps),
#				 - ClusteringCheck (Verify the coherency of all parameters for the clustering),
#				 - CheckRawDataCoherency (Check the coherency of steps and files given by the user),
#				 - CheckNamesUnicity (Verify the names and MIDs of all samples),
#				 - CheckHomogenizationChoice (Check the homogenization step),
#				 - CheckClusteringCoherency (Verify if the alignment is done for clustering steps),
#				 - CheckHuntingCoherency (Verify if essential steps are done for the hunting),
#				 - CheckRecoveringCoherency (Verify if essential steps are done for the recovering),
#				 - CheckComputationAnalysisCoherency (Verify if essential steps are chosen),
#				 - CheckGlobalAnalysisCoherency (Verify if essential steps are done for the analysis),
#				 - CheckUniFracAnalysisCoherency (Verify if essential steas are done for the analysis),
#				 - VerifyingUserChoices (main function lauching and testing all choices and steps).
#
#	   DETAILS:  1.1: We modified the functions "VerifyingUserChoices" and "CheckRawDataCoherency" to
#				 integrate the use of fastq files from Illumina sequencing systems. Done the 09/09/2014.
#
# REQUIREMENTS:  Perl language, and Math library to compute binary files
#       AUTHOR:  Sebastien TERRAT <sebastien.terrat@dijon.inra.fr>
#      COMPANY:  UMR 1347 Agroecology - Platform GenoSol
#      VERSION:  1.1
#    REALEASED:  17/04/2014
#     REVISION:  09/09/2014
#=======================================================================================================

# Name of the package defined for the main program
package PerlLib::VerifyingUserChoices;

# Needed libraries
use strict;
use warnings;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function to print the information for the user as all tests have been passed for the evaluated step.
# It takes the name of the evaluated step as input.
sub InfoForUser
{
	return "\n#############################################################\nChosen names for the $_[0] step have been checked, 
and all validation steps have been passed.\n#############################################################\n";
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function to check if the global alignment step (RAW or CLEAN) is realized, as this step is essential
# for many steps in GnS-PIPE: like the clustering or the computation ones. It takes as input the name of
# alignment needed to be verified, and the step depending of the alignment. 
sub AlignmentCheck
{
	# Needed variables
	my $step = $_[0];
	my $verified_step = $_[1];
	my $ref_array = "";
	my @parameters = ();
	
	# We catch the alignment parameters chosen by the user
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters($_[0]);
	@parameters = @$ref_array;
	# If the step is not realized
	if(uc($parameters[0]) eq "NO")
	{
		die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
$verified_step step, however the $step was not realized. So, the analysis
can't be realized. Please, correct this error in the Input.txt file and relaunch GnS-PIPE.\n";
	}
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function to check if the clustering step (RAW or CLEAN) is realized, and if the chosen dissimilarity
# percentage is computed based on user parameters. It takes as input the name of the alignment needed to
# be verified, the percentage chosen for the hunting step and the step name depending of the clsutering.
sub ClusteringCheck
{
	# Needed variables
	my $step = $_[0];
	my $chosen_percentage = $_[1];
	my $verified_step = $_[2];
	my $ref_array = "";
	my @parameters = ();
	my @clust_steps = ();
	my $clust_verif = 0;
	
	# We catch the clustering parameters
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters($step);
	@parameters = @$ref_array;
	# If the step is not realized
	if(uc($parameters[0]) eq "NO")
	{
		die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision 
to launch the $verified_step step, however the $step was not 
realized. So, the global analysis can't be realized.  Please, correct this
error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	# If the step is chosen by the user
	else
	{
		# We verify the existence of the same percentage chosen for the clustering and the hunting
		$ref_array = PerlLib::Miscellaneous::ComputeClusteringSteps($parameters[3], $parameters[4]);
		@clust_steps = @$ref_array;
		foreach (@clust_steps) { if ($chosen_percentage == $_) { $clust_verif = 1; } }
		if ($clust_verif == 0)
		{
			die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision to 
launch the $verified_step step, however the chosen percentage of dissimilarity
in the $verified_step step ($chosen_percentage%) will not be computed based on $step
chosen parameters. So, the global analysis can't be realized. 
Please, correct this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
		}
	}
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function developed to verify the coherency between the files stored in the Raw_data folder, and the
# steps chosen by the user, as it can be impossible to realize some steps (like the #QUALITY_CLEANING one
# if the .qual files are not stored in the Raw_data folder). It takes as input three arguments (the list
# of .sff files, the list of .qual files and the list of .fna/.fasta files.
sub CheckRawDataCoherency
{
	# Needed variables
	my $ref_array = "";
	my $step = "";
	my @parameters = ();
	my @files_sff = @{$_[0]};
	my @files_qual = @{$_[1]};
	my @files_fna = @{$_[2]};
	my @files_fastq = @{$_[3]};
	my @steps_qual = ("EVAL_QUAL", "QUALITY_CLEANING");
	my @steps_fna = ("PREPROCESS_MIDS", "RANDOM_PREPROCESS_MIDS", "PREPROCESSING", "RANDOM_PREPROCESS", 
"RAW_INFERNAL_ALIGNMENT", "RAW_CLUSTERING", "HUNTING", "RECOVERING", "RANDOM_CLEANED", 
"TAXONOMY", "CLEAN_INFERNAL_ALIGNMENT", "CLEAN_CLUSTERING", "COMPUTATION", "GLOBAL_ANALYSIS", 
"UNIFRAC_ANALYSIS");
	
	# We first verify if we have .sff or .fastq files, iy it's the case, we can do every step
	if($#files_sff > -1) { return "OK"; }
	elsif($#files_fastq > -1) { return "OK"; }
	
	# Loop to test each step where a quality file can be implied and used
	foreach $step (@steps_qual)
	{
		$ref_array = PerlLib::Miscellaneous::SearchInputParameters($step);
		@parameters = @$ref_array;
		if(uc($parameters[0]) eq "YES")
		{
			if($#files_qual == -1)
			{
				die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision to
realize the $step step, but you did not put in the Raw_data folder a .qual file. Please 
correct this error (in the Input.txt file or in the Raw_data folder) and relaunch GnS-PIPE.\n";
			}
		}
	}
	# Loop to test each step where a .fasta and/or a .fna file can be implied and used
	foreach $step (@steps_fna)
	{
		$ref_array = PerlLib::Miscellaneous::SearchInputParameters($step);
		@parameters = @$ref_array;
		if(uc($parameters[0]) eq "YES")
		{
			if($#files_fna == -1)
			{
				die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision to
realize the $step step, but you did not put in the Raw_data folder a .fna file. Please 
correct this error (in the Input.txt file or in the Raw_data folder) and relaunch GnS-PIPE.\n";
			}
		}
	}
	return "\n#############################################################\nChosen steps coherency have been checked using the given raw 
files, and all validation steps have been passed.\n#############################################################\n\n";
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function developed to verify if the names given by the user for samples, but also MIDs and raw files
# are correct. It uses the function &SearchInputParameters to catch needed information and check the 
# choices of the user.
sub CheckNamesUnicity
{
	# Needed variables
	my $ref_array = "";
	my $filename = "";
	my $i = 0;
	my @parameters = ();
	my @tmp = ();
	my %filenames = ();
	my %samples = ();
	my %MIDs = ();
	
	# First, we catch all parameters for the PREPROCESS_MIDS step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("PREPROCESS_MIDS");
	@parameters = @$ref_array;
	# We avoid the analysis if the step is not used by the user
	if(uc($parameters[0]) eq "NO") { return "OK"; }
	
	# We first catch the names of the filenames
	LOOP01:for($i=1;$i<=$#parameters;$i++)
	{
		@tmp = split("/", $parameters[$i]);
		$filenames{$tmp[0]}++;
	}
	# We first create %samples and %MIDs to check their unicity
	LOOP02:foreach $filename (keys(%filenames))
	{
		# For each line found
		LOOP03:for($i=1;$i<=$#parameters;$i++)
		{
			if($parameters[$i] =~ m/$filename/)
			{
				# We split the line to separate the filename, the MID and the sample name
				@tmp = split("/", $parameters[$i]);
				# We store needed information in dedicated hash tables
				$MIDs{$tmp[1]}++;
				$samples{$tmp[2]}++;
			}
		}
		# We verify the unicity of %samples
		LOOP04:foreach (keys(%samples))
		{
			if ($samples{$_} > 1)
			{
				die "==> The GnS-PIPE was stopped due to an unicity error. Your sample named: $_
was present at least two times for the same chosen filename: $filename".".
Please correct this error in the Input.txt file and relaunch GnS-PIPE\n";
			}
		}
		# We verify the unicity of %MIDs
		LOOP05:foreach (keys(%MIDs))
		{
			if ($MIDs{$_} > 1)
			{
				die "==> The GnS-PIPE was stopped due to an unicity error. Your MID named: $_
was present at least two times for the same chosen filename: $filename".".
Please correct this error in the Input.txt file and relaunch GnS-PIPE.\n";
			}
		}
		# We emptied the hash tables to analyze a second potential filename
		%MIDs = ();
		%samples = ();
	}
	# We finally send back information to the user
	&InfoForUser("PREPROCESS_MIDS");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function needed to determine efficiently where the concatenation of files will be realized to 
# to the global analysis step.
sub CheckHomogenizationChoice
{
	# Needed variables
	my $ref_array = "";
	my $step = "";
	my $verif_rdm = 0;
	my @parameters = ();
	my @potential_steps = ("RANDOM_PREPROCESS_MIDS", "RANDOM_PREPROCESS", "RANDOM_CLEANED");
	
	# Loop to test the user choices for all potential steps of homogenization 
	LOOP01:foreach $step (@potential_steps)
	{
		$ref_array = PerlLib::Miscellaneous::SearchInputParameters($step);
		@parameters = @$ref_array;
		if (uc($parameters[0]) eq "YES") { $verif_rdm++; }
	}
	# We indicate to the user that his selction is wrong, before stopping the complete program
	if ($verif_rdm > 1)
	{
		die "==>You can't realize several random selection during the analysis,\nas you will bias your data !! 
		You must modify your choices in the Input.txt file and try again.\n";
	}
	else { return $verif_rdm; }
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function checking if the clustering step (RAW or CLEAN) is realized, as this step is essential for
# many steps in GnS-PIPE: like the clustering or the computation ones. It takes as input the name of the
# checked clustering  step.
sub CheckClusteringCoherency
{
	# Needed variables
	my $clust_step = $_[0];
	my $ref_array = "";
	my @parameters = ();
	
	# First, we verify if the user chose to do a clustering step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters($clust_step);
	@parameters = @$ref_array;
	if (uc($parameters[0]) eq "NO") { return "NO"; }
	
	# We also check if the alignment is done, as it is essential for the clustering
	if($clust_step eq "RAW_CLUSTERING") { &AlignmentCheck("RAW_INFERNAL_ALIGNMENT", $clust_step); }
	elsif($clust_step eq "CLEAN_CLUSTERING") { &AlignmentCheck("CLEAN_INFERNAL_ALIGNMENT", $clust_step); }
	#Finally, we send back information to the user
	&InfoForUser($clust_step);
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function to check if the hunting step is realized with all needed previous steps, as this step needs
# a clustering and an alignment step.
sub CheckHuntingCoherency
{
	# Needed variables
	my $ref_array = "";
	my @parameters = ();
	
	# First, we verify if the user chose to do a hunting step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("HUNTING");
	@parameters = @$ref_array;
	if (uc($parameters[0]) eq "NO") { return "NO"; }
	
	# We check for all essential steps for the hunting.
	&AlignmentCheck("RAW_INFERNAL_ALIGNMENT", "HUNTING");
	&ClusteringCheck("RAW_CLUSTERING", $parameters[1], "HUNTING");
	&InfoForUser("HUNTING");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function to check if the recovering step is realized with all needed previous steps, as the RECOVERING
# step needs a clustering, an alignment step, but also the hunting step.
sub CheckRecoveringCoherency
{
	# Needed variables
	my $ref_array = "";
	my @parameters = ();
	my @parameters_recovering = ();
	
	# First, we verify if the user chose to do a recovering step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("RECOVERING");
	@parameters_recovering = @$ref_array;
	if (uc($parameters_recovering[0]) eq "NO") { return "NO"; }
	# Then, we verify if the user chose a good organism (bacteria or fungi)
	if ((uc($parameters_recovering[1]) ne "BACTERIA")&&(uc($parameters_recovering[1]) ne "FUNGI"))
	{
		die "\n==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
RECOVERING step, however the choice of the organism ($parameters_recovering[1]) was not coherent. You must 
chose between 'bacteria' or 'fungi' So, the analysis can't be realized. Please, correct
this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	# Then, we verify if the user chose a good database (RDP or Silva)
	if ((uc($parameters_recovering[2]) ne "RDP")&&(uc($parameters_recovering[2]) ne "SILVA"))
	{
		die "\n==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
RECOVERING step, however the choice of the database ($parameters_recovering[2]) was not coherent. You must 
chose between 'RDP' or 'Silva' So, the analysis can't be realized. Please, correct
this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	# Finally, we check if the user chose an available taxonomic level (domain, phylum, class, order,
	# family or genus)
	if ((uc($parameters_recovering[3]) ne "DOMAIN")&&(uc($parameters_recovering[3]) ne "PHYLUM")
	&&(uc($parameters_recovering[3]) ne "CLASS")&&(uc($parameters_recovering[3]) ne "ORDER")
	&&(uc($parameters_recovering[3]) ne "FAMILY")&&(uc($parameters_recovering[3]) ne "GENUS"))
	{
		die "\n==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
RECOVERING step, however the choice of the taxonomic level ($parameters_recovering[3]) was not coherent. 
You must chose between [domain, phylum, class, order, family or genus]. So, the analysis 
can't be realized. Please, correct this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	
	# Then, we verify the coherency between the chosen steps of the user
	&AlignmentCheck("RAW_INFERNAL_ALIGNMENT", "RECOVERING");
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("RAW_INFERNAL_ALIGNMENT");
	@parameters = @$ref_array;
	if(uc($parameters_recovering[1]) ne uc($parameters[1]))
	{
		die "\n==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
RECOVERING step, however the model of the RAW_INFERNAL_ALIGNMENT ($parameters[1]) was not coherent.
So, the analysis can't be realized. Please, correct this error in the Input.txt file 
and relaunch GnS-PIPE.\n\n";
	}
	
	# Finally, we check if the hunting step is done
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("HUNTING");
	@parameters = @$ref_array;
	if (uc($parameters[0]) eq "NO")
	{
		die "\n==> The GnS-PIPE was stopped due to a coherency error. You take the decision to 
launch the RECOVERING step, however the HUNTING was not realized. So, the analysis
can't be realized. Please, correct this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	# We give back information to the user
	&InfoForUser("RECOVERING");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function developed to verify if parameters chosen by the user were coherent: more particularly, if the
# clustering level chosen for the computation step was used during the previous clustering step.
sub CheckComputationAnalysisCoherency
{
	# Needed variables
	my $ref_array = "";
	my @parameters = ();
	
	# We catch the computation parameters
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("COMPUTATION");
	@parameters = @$ref_array;
	
	# We verify that the user chose to do the computation step, or we leave the verification step
	if(uc($parameters[0]) eq "NO") { return "NO"; }
	
	# We then verify using the ClusteringCheck function if the clusetring is realized and if
	# parameters are correct.
	if((uc($parameters[1]) eq "RAW")||(uc($parameters[1]) eq "BOTH"))
	{
		&AlignmentCheck("RAW_INFERNAL_ALIGNMENT", "COMPUTATION");
		&ClusteringCheck("RAW_CLUSTERING", $parameters[2], "COMPUTATION");
	}
	if((uc($parameters[1]) eq "CLEAN")||(uc($parameters[1]) eq "BOTH"))
	{
		&AlignmentCheck("CLEAN_INFERNAL_ALIGNMENT", "COMPUTATION");		
		&ClusteringCheck("CLEAN_CLUSTERING", $parameters[2], "COMPUTATION");
	}
	# We give back information to the user
	&InfoForUser("COMPUTATION");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function developed to verify the coherency of all parameters chosen by the user for the global 
# analysis step, like if the global alignment or the taxonomy steps are done if needed.
sub CheckGlobalAnalysisCoherency
{
	# Needed variables
	my $ref_array = "";
	my @parameters = ();
	my @parameters_glob = ();
	
	# We first catch and store chosen parameters fro the global analysis step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("GLOBAL_ANALYSIS");
	@parameters_glob = @$ref_array;
	# We verify that the user chose to do the global analysis step, or we leave the verification step
	if(uc($parameters_glob[0]) eq "NO") { return "NO"; }
	
	# We first verify if the alignment was done (RAW).
	if((uc($parameters_glob[1]) eq "RAW")||(uc($parameters_glob[1]) eq "BOTH"))
	{
		&AlignmentCheck("RAW_INFERNAL_ALIGNMENT", "GLOBAL_ANALYSIS");	
		&ClusteringCheck("RAW_CLUSTERING", $parameters_glob[2], "GLOBAL_ANALYSIS");
	}
	# We also verify if the alignment was done (CLEAN).
	if((uc($parameters_glob[1]) eq "CLEAN")||(uc($parameters_glob[1]) eq "BOTH"))
	{
		&AlignmentCheck("CLEAN_INFERNAL_ALIGNMENT", "GLOBAL_ANALYSIS");			
		&ClusteringCheck("CLEAN_CLUSTERING", $parameters_glob[2], "GLOBAL_ANALYSIS");	
	}
	
	# We finally verify if the taxonomy was done as the user can need it in the global analysis step.
	if(uc($parameters_glob[4]) eq "YES")
	{
		# We catch the taxonomy parameters
		$ref_array = PerlLib::Miscellaneous::SearchInputParameters("TAXONOMY");
		@parameters = @$ref_array;
		if(uc($parameters[0]) eq "NO")
		{
			die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision to launch the 
GLOBAL_ANALYSIS step, however the TAXONOMY was not realized. As you chose to assign the 
OTUs of the global analysis, you need the TAXONOMY step. Please, correct this error in
the Input.txt file and relaunch GnS-PIPE.\n\n";
		}
	}
	# We give back information to the user
	&InfoForUser("GLOBAL_ANALYSIS");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function developed to verify if parameters chosen by the user were coherent: more particularly, if the
# phylogenetic tree was computed for the UniFrac analysis.
sub CheckUniFracAnalysisCoherency
{
	# Needed variables
	my $ref_array = "";
	my @parameters_unifrac = ();
	my @parameters_glob = ();
	
	# We first catch and store chosen parameters fro the global analysis step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("UNIFRAC_ANALYSIS");
	@parameters_unifrac = @$ref_array;
	
	# We verify that the user chose to do the UniFrac analysis step, or we leave the verification step
	if(uc($parameters_unifrac[0]) eq "NO") { return "NO"; }
	
	#We then catch and store chosen parameters for the global analysis step
	$ref_array = PerlLib::Miscellaneous::SearchInputParameters("GLOBAL_ANALYSIS");
	@parameters_glob = @$ref_array;
	# If chosen parameters are incoherent, we stop the program, and end back information to the user
	if(uc($parameters_glob[6]) eq "NO")
	{
		die "==> The GnS-PIPE was stopped due to a coherency error. You take the decision 
to launch the UNIFRAC_ANALYSIS step, however the GLOBAL_ANALYSIS phylogenetic 
tree will not be realized based on your parameters. 
Please, correct this error in the Input.txt file and relaunch GnS-PIPE.\n\n";
	}
	# We give back information to the user that everything is OK
	&InfoForUser("UNIFRAC_ANALYSIS");
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#=======================================================================================================
#=======================================================================================================
# MAIN FUNCTION CALLED BY THE MAIN PROGRAM
#=======================================================================================================
#=======================================================================================================
# FILE:			VerifyingUserChoices
# USAGE:		PerlLib::VerifyingUserChoices::VerifyingUserChoices(ARGUMENTS)
# ARGUMENTS:	- $_[0]: Reference to the @files_sff list listing all raw .SFF files found,
#				- $_[1]: Reference to the @files_qual list listing all raw .QUAL files found,
#				- $_[2]: Reference to the @files_fna list listing all raw .FNA/.FASTA files found,
#
# DESCRIPTION:	This function was dedicated to launch all steps and functions dedicated to check
#				the coherency between the user choices and the steps and dependencies of the steps. It
#				takes as input three lists, to verify is needed files are given by the user, and if 
#				these files are coherent with his choices.
#=======================================================================================================
sub VerifyingUserChoices
{
	# Input variables
	my $ref_files_sff = $_[0];
	my $ref_files_qual = $_[1];
	my $ref_files_fna = $_[2];
	my $ref_files_fastq = $_[3];
	
	# Needed variables
	my $validation_text = "";
	my $verif_rdm = "";
	
	# We check the coherency between chosen steps (EVAL_QUAL and QUALITY_CLEANING for example) and raw data
	# given by the user
	$validation_text = &CheckRawDataCoherency($ref_files_sff, $ref_files_qual, $ref_files_fna, $ref_files_fastq);
	if($validation_text ne "OK") { print $validation_text; };

	# We check the unicity of sample names and chosen MIDs given by the user for each raw file.
	$validation_text = &CheckNamesUnicity();
	if($validation_text ne "OK") { print $validation_text; };

	# We check the coherency for the homogenization step.
	$verif_rdm = &CheckHomogenizationChoice();
	if ($verif_rdm == 0)
	{
		print "==> You chose to not homogenize your sequence sets. <==
If a GLOBAL_ANALYSIS step is done, the treatment will be realized on:
- the cleaned reads (after hunting and/or recovering steps if they are realized),
- or at least on raw reads.\n";
	}
	# We check the coherency for the raw clustering step.
	$validation_text = &CheckClusteringCoherency("RAW_CLUSTERING");
	if($validation_text ne "NO") { print $validation_text; };
	# We check the coherency for the hunting step.
	$validation_text = &CheckHuntingCoherency();
	if($validation_text ne "NO") { print $validation_text; };
	# We check the coherency for the recovering step.
	$validation_text = &CheckRecoveringCoherency();
	if($validation_text ne "NO") { print $validation_text; };
	# We check the coherency for the clean clustering step.
	$validation_text = &CheckClusteringCoherency("CLEAN_CLUSTERING");
	if($validation_text ne "NO") { print $validation_text; };
	# We check all parameters for the global analysis step, as this particular part of GnS-PIPE would need
	# some results given by previous steps (e.g. clustering, or taxonomy)
	$validation_text = &CheckComputationAnalysisCoherency();
	if($validation_text ne "NO") { print $validation_text; };
	# We check all parameters for the global analysis step, as this particular part of GnS-PIPE would need
	# some results given by previous steps (e.g. clustering, or taxonomy)
	$validation_text = &CheckGlobalAnalysisCoherency();
	if($validation_text ne "NO") { print $validation_text; };
	# We check all parameters for the global analysis step, as this particular part of GnS-PIPE would need
	# some results given by previous steps (e.g. clustering, or taxonomy)
	$validation_text = &CheckUniFracAnalysisCoherency();
	if($validation_text ne "NO") { print $validation_text; };
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

1;
