#=======================================================================================================
#         FILE:   	perl Fungi_SILVA_Taxonomy_3.5.pl
#        USAGE:   	perl Fungi_SILVA_Taxonomy_3.5.pl
#  DESCRIPTION:   	This PERL program was developed to launch in parallel a BLASTn analysis of each read
#				against the SILVA database of fungi 18S downloaded the 06/09/2012 (release 111 on 
#				May 2012). This database contain more then 54756 sequences of fungi 18S.
#				
#				Several developements were done (details are in the program itself). Briefly, we
#				first launch a MEGABLAST step on the fungal database from SILVA already formatted. 
#				The BLAST results are temporary stored in the Temporary folder, and then treated to
#				create the needed file stored in the dedicated folder. The treatment of BLAST results
#				is based first on the identity of the sequences, then on the frequency of results, and
#				finally on the information linked to the results (we, for example, favor a lower number
#				of sequences more informative than Environmental sequences).
#				Three different more names have been created: Environmental (sequences found in the 
#				database, but not annotated), Unclassified (sequences found in the database, but with
#				no information at this level), and Unknown (no sequences with results).
#
#				Modifs 3.1 (05/10/2012): 
#
#				Modifs 3.5 (18/06/2013): 
#				
#       AUTHOR:   	Sebastien TERRAT <sebastien.terrat@dijon.inra.fr>
#      COMPANY:   	UMR 1347 Agroecology, GenoSol platform
#      VERSION:   	3.5
#    REALEASED:   	23/01/2012
#     REVISION:   	18/06/2013
#=======================================================================================================
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#=======================================================================================================

#!/usr/bin/perl
#Print errors and warnings during the implementation of the perl program
use strict; use warnings;
use FindBin;

#=======================================================================================================
###SUMMARY:	Function needed to analyze obtained BLAST results at a defined level given by the user.
#For each element of the given array (here $_[0]:@Taxo_id_OK), we will analyze BLAST results to find the 
#best information at the given level, based on various criteria, like their frequency, their informative
#level, etc... All details can be found in the function itself.

###INPUT:		A reference ot an array (here $_[0]: @Taxo_id_OK), a string to define the analyzed level
#for example: "family", and the string used to store the results(here$_[2]: $result_tmp).
###OUTPUT:	A string concatenating previous results given as input and the new results, giving the 
#level name found and the identity percentage in form.
sub AnalyzeBlast_result
{
	#Defined local empty variables
	my $average = 0;				my $result_in_function = "";		my $result = "";
	my $number_found = 0;			my @tmp = ();					my $env_check = 0;
	my @Infos = ();				my @Names = ();				my $env_i = 0;
	my ($i,$j,$k,$l) = (0,0,0,0);		my $Nb_names = 0;
	
	#We first evaluate the length of our results, to avoid the analysis of empty arrays
	$i = scalar(@{$_[0]});
	#If the array is empty (no results available at the defined threshold)
	if($i == 0)
	{
		#We create the result that then will be returned
		if ($_[1] eq "kingdom") {$result_in_function = "Unknown\tdomain\t0.00\t".$_[2];}
		else {$result_in_function = "Unknown\t$_[1]\t0.00\t".$_[2];}
		#We return it and stop the Function
		return $result_in_function;
	}
	#Here, we defined $i at 0 as needed
	else {$i = 0;}
	
	#We use this loop to find each name of the taxonomic level found and regroup them in the array @Infos to 
	#count the occurrences of each name and the number of [C] sequences for each one.
	
	#NOTE: we need here to use $_[0] as a clear reference of an array with the @{} element to indicate that we use
	#the $_[0] reference as an array (references subroutine to see details).
	LOOP1:foreach $result (@{$_[0]})
	{
		#We first catch the name of the level
		if ($result =~ m/\;(\w+)\($_[1]\);/)
		{
			#We store it in the hash table to count the number of occurences of this name in the analyzed
			#BLAST result
			#We store also this information in the @Infos array
			$Infos[$i][0] = $1;
		}
		#If the BLAST result is an environmental sequence isolated, but not assigned
		elsif ($result =~ m/\[E\]/)
		{
			#We store "Environmental" to create a new taxonomic information (as this sequence exist in 
			#other studies, but was not affiliated).
			$Infos[$i][0] = "Environmental";
		}
		#If the level is not found in the BLAST result
		else
		{
			#If we didn't found the name (we indicate the status "Unclassified")
			$Infos[$i][0] = "Unclassified";
		}
		#We then split the line and compute the identity value
		@tmp = split("\t", $result);
		$Infos[$i][1] = ($tmp[5]-$tmp[6])/$tmp[1];
		#We store also the information that the BLAST affiliation information is Complete, Incomplete
		#or from an Environment
		$result =~ m/\[(.*)\]/;
		$Infos[$i][2] = $1;
		$i++;
	}
	
	#We take the names of the BLAST results and store them in the array @Names with other details
	$Names[0][0] = $Infos[0][0];
	#We put 0 as number of occurences and [C] sequences for this name
	$Names[0][1] = 0;	$Names[0][2] = 0;
	
	#Here, we read all elements of @Names to analyze them and store summaries in @Infos
	#Loop to read all names found for all BLAST results
	LOOP2:for($j=0;$j<$#Infos+1;$j++)
	{
		#Loop to compare the chosen name to all stored names in the array @Names
		LOOP3:for($k=0;$k<=$Nb_names;$k++)
		{
			#If the name is already stored
			if($Names[$k][0] eq $Infos[$j][0])
			{
				#We increase the number of occurences
				$Names[$k][1]++;
				if ($Infos[$j][2] eq "C") {$Names[$k][2]++;}
				#We go for the next names needed analysis
				next LOOP2;
			}
		}
		#If the name is not found in the @Names array, we create it
		$Names[$Nb_names+1][0] = $Infos[$j][0];
		#We count the value 1 as its the first occurrence
		$Names[$Nb_names+1][1] = 1;
		#We also verify the [C] existence of sequences
		if ($Infos[$j][2] eq "C") {$Names[$k][2] = 1;}
		else {$Names[$k][2] = 0;}
		#We increase the number of names found
		$Nb_names++;
	}	
	$i = 0; $result = "";
	#Loop to read all @Names with the count of occurences and [C] sequences
	LOOP4:for($k=0;$k<$#Names+1;$k++)
	{
		#We search the name with the maximum of occurences
		if ($Names[$k][1] > $i)
		{
			#We store it and the corresponding occurence
			$result = $Names[$k][0];
			#We store also the number of occurences
			$i = $Names[$k][1];
			#Finally, we store also its position in the array @Names
			$l = $k;
		}
	}
	
	#Now, we will verify that we can choose only the "Environmental" name, or we will favor 
	#another choice potentially available in the list @Infos
	if ($result eq "Environmental")
	{
		#We first check for the existence of other possibilities using the length of the 
		#raw BLAST results stored in the array given as input
		if ($i < scalar(@{$_[0]}))
		{
			#We store for analysis $env_i and $env_check
			$env_check = 1; $env_i = $i; $i = 0;
			##Loop to read all @Names with the count of occurences and [C] sequences
			LOOP5:for($k=0;$k<$#Names+1;$k++)
			{
				#We search the name with the maximum of occurrences not Environmental and not Unknown
				if (($Names[$k][1] > $i)&&($Names[$k][0] ne "Environmental")&&($Names[$k][0] ne "Unknown"))
				{
					#We store it and the corresponding occurence
					$result = $Names[$k][0];
					#We store also the number of occurences
					$i = $Names[$k][1];
					#Finally, we store also its position in the array @Names
					$l = $k;
				}
			}
		}
	}
	
	#Loop to verify the existence of two different names with the same number of
	#occurences
	LOOP6:for($k=0;$k<$#Names+1;$k++)
	{
		#If we found two names, we first verify that it's not the same detected
		if (($Names[$k][1] == $i) && ($result ne $Names[$k][0]))
		{
			#If the name is an Environmental group, we chose the second one
			if ($result eq "Environmental") {$result = $Names[$k][0]; $i = $Names[$k][1];}
			#If it's not an Environmental group, we will choose the name with 
			#the maximum of [C] sequences as the information if valuable
			else
			{
				#We test the frequency of [C] sequences
				if ($Names[$k][2] > $Names[$l][2]) {$result = $Names[$k][0]; $i = $Names[$k][1];}
				#If we can't differentiate the two names, we decide to return an Unknown result
				elsif ($Names[$k][2] == $Names[$l][2])
				{
					#As we can consider the possibility that the maximum was "Environmental"
					#and the two other below possibilities impossible to differentiate, we keep
					#the "Environmental" choice
					if ($env_check == 1) {$result = "Environmental"; $i = $env_i;}
					#If not, we must indicate the Unknown choice
					else
					{
						#We create the result then we return it
						if ($_[1] eq "kingdom") {$result_in_function = "Unknown\tdomain\t0.00\t".$_[2];}
						else {$result_in_function = "Unknown\t$_[1]\t0.00\t".$_[2];}
						return $result_in_function;
					}
				}
			}
		}
	}	
	
	#Last loop to define the percentage of identity based on the consensus of the chosen name
	LOOP7:for($j=0;$j<$#Infos+1;$j++)
	{
		#Test to find the identity for each BLAST result, and we add the find score to $average
		if ($result eq $Infos[$j][0]) {$average = $average + $Infos[$j][1];}
	}
	#We then compute the average value using the number of occurrences found
	$average = sprintf("%0.2f", $average/$i);
	#We finally return the needed information
	if ($_[1] eq "kingdom") {$result_in_function = "$result\tdomain\t$average\t".$_[2];}
	else {$result_in_function = "$result\t$_[1]\t$average\t".$_[2];}
	return $result_in_function;
}
# End of the function
#=======================================================================================================


#=======================================================================================================
###SUMMARY: 	Function needed to define if BLAST results are analyzed or not (based on their presence 
#in arrays @Taxo_id_OK and @Taxo_id_NOT) at a defined threshold for each taxonomic level.
#
###INPUT: 	two references of arrays (here $_[1]: @Taxo_id_NOT and $_[0]: @Taxo_id_OK) 
#and the threshold for each level ($_[2]).
###OUPUT:		NONE

sub DefineBlast_analyzed
{
	#Defined local variables
	my ($i, $identity) = (0,0);	my $line = "";		my @line = ();
	#For each element of BLAST results not yet used based on their identity scores
	foreach $line (@{$_[1]})
	{
		#We split the BLAST output line
		@line = split("\t", $line);
		#We compute the identity 
		$identity = ($line[5]-$line[6])/$line[1];
		#If the identity is higher than the given threshold
		if($identity > $_[2])
		{
			#We delete the line in the @Taxo_id_NOT to store it in @Taxo_id_OK
			@line = splice(@{$_[1]},$i,1);
			push(@{$_[0]}, $line);
		}
		#We count each line
		$i++;
	}
}
# End of the function
#=======================================================================================================

#=======================================================================================================
# MAIN PROGRAM
#=======================================================================================================

#Defined main variables
my $i = 0;
my $identity = 0;
my $hits = 0;
my $taxo_exec = "";
my $line = "";
my $result_final = "";
my $result_tmp = "";
my @line = ();
my @Taxo_id_OK = ();
my @Taxo_id_NOT = ();

my $folder_Input = $ARGV[0];
my $folder_Output = $ARGV[1];
my $file = $ARGV[2];
my $cores = $ARGV[3];
my $threshold = $ARGV[4]/100;
#3.5: We modify this program to catch the same database as others in the dedicated folder of GnS-PIPE
my $folder_DB = "$FindBin::RealBin/../Data/Databases/FUNGI/SILVA/";



#=======================================================================================================
#First step, we launch the blast step using the information given by the Taxo_recovery_2.0.pl program
#(folder, and file used for the analysis).
#=======================================================================================================

#3.5: Launch the program will all needed parameters ($folder_Input, $file, other parameters)
$taxo_exec = "blastn -query Result_files/$folder_Input/$file -strand plus -task megablast -db $folder_DB/arb-silva.de_2013-29-04_id56261_vf.fasta -out Temporary_files/Blast_Tmp_$file -outfmt \"7 qseqid qlen qstart qend evalue nident gaps sseqid\" -evalue 1e-50 -word_size 7 -max_target_seqs 15 -num_threads $cores";
system($taxo_exec);

#=======================================================================================================
#Second step, analyze all temporary result files to create the needed file in the OUT/ directory
#=======================================================================================================

#Opening the needed files (the temporary file and the out file)
open (F_TMP, "Temporary_files/Blast_Tmp_$file")|| die "==> File Not Found: Temporary_files/Blast_Tmp_$file<==";
open (F_RESULTS, ">Result_files/$folder_Output/Taxo_SILVA_$file")|| die "==> Can't create file: Taxo_SILVA_$file<==";
LOOP3:while (<F_TMP>)
{
	#Catch the information fo the read (based on the information line # Query:
	if ($_ =~ m/# Query: (.+)\n/) {$result_final = $result_final."$1\t\tRoot\tnorank\t1.0\t";}
	#Catch the number of hits found (between 0 and 15, particular cases more)
	if ($_ =~ m/# (\d+) hits found/)
	{
		$hits = $1;
		if ($hits == 0)
		{
			$result_final = $result_final."Unknown\tdomain\t0.00\tUnknown\tphylum\t0.00\tUnknown\tsubphylum\t0.00\tUnknown\tclass\t0.00\tUnknown\tsubclass\t0.00\tUnknown\torder\t0.00\tUnknown\tsuborder\t0.00\tUnknown\tfamily\t0.00\tUnknown\tsubfamily\t0.00\tUnknown\tgenus\t0.00\t";
			printf (F_RESULTS "$result_final\n");
			$result_final = "";
			next LOOP3;
		}
		#As we need only five results, but find more due to subsequences and subalignement, need to modify it.
		if ($hits > 15) {$hits = 15;}
		#Based on this number ($hits), we read each line needed	
		for($i=0;$i<$hits;$i++)
		{		
			$line = <F_TMP>;
			@line = split("\t", $line);
			$identity = ($line[5]-$line[6])/$line[1];			
			#For genus analysis of fungi, if this value is below 0.97, we reject the read and kept it in the array @Taxo_id_NOT
			if($identity > $threshold) {push(@Taxo_id_OK, $line);}
			else {push(@Taxo_id_NOT,$line);}
		}
		#We launch the defined function &AnalyzeBlast_result for the genus treatment
		#NOTE: here, we defined \@Taxo_id_OK (a reference for the array @Taxo_id_OK, not the array itself), to simplify its using.
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "genus", $result_tmp);
		
		#For family and subfamily analysis of fungi, if this value is below 0.94, we reject the read and kept it in the array @Taxo_id_NOT
		#We use here the second function names DefineBlast_analyzed to add BLAST results to @Taxo_id_OK and then the first function
		#&AnalyzeBlast_result to analyze our results
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "subfamily", $result_tmp);
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "family", $result_tmp);
		#For order analysis of fungi, if this value is below 0.90, we reject the read and kept it in the array @Taxo_id_NOT
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "suborder", $result_tmp);
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "order", $result_tmp);
		#For class analysis of fungi, if this value is below 0.85, we reject the read and kept it in the array @Taxo_id_NOT
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "subclass", $result_tmp);
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "class", $result_tmp);
		#For phylum analysis of fungi, if this value is below 0.80, we reject the read and kept it in the array @Taxo_id_NOT
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "subphylum", $result_tmp);
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "phylum", $result_tmp);
		#For domain analysis of fungi, if this value is below 0.80, we reject the read and kept it in the array @Taxo_id_NOT
		&DefineBlast_analyzed(\@Taxo_id_OK, \@Taxo_id_NOT, $threshold);
		$result_tmp = &AnalyzeBlast_result(\@Taxo_id_OK, "kingdom", $result_tmp);		
		printf (F_RESULTS $result_final.$result_tmp."\n");

		#Needed empty variables for the next read analyzed
		@Taxo_id_OK = ();@Taxo_id_NOT= ();
		$result_final = ""; $result_tmp = "";
	}
}
close F_RESULTS;
#We finally delete the Temporary files not needed anymore.
unlink("Temporary_files/Blast_Tmp_$file");
