#=======================================================================================================
#         FILE:  Infernal_align_2.7.pl
#        USAGE:  perl Infernal_align_2.7.pl
#  DESCRIPTION:  This program was developed to efficiently parallelize the INFERNAL aligner (1.0.2) 
#			  defined by EP Nawrocki, DL Kolbe and SR Eddy (http://selab.janelia.org). As INFERNAL 
#			  can be run by Message Passing Interface (MPI) but seems to be not functional with our
#			  computational server, we decided to parallelize the alignment step by parsing sequence
#			  files. For this, we let the user choose the number of cores he want to use, but also the
#			  choice of the used model (bacteria or archaea) and if he want to merge all alignments 
#			  of treated sequences files.
#			  All sequence files needed alignments are stored in the IN/ directory. The Matrix/ one
#			  store the models and their alignements used to design them. The Tmp/ directory is used 
#			  to store all temporary files (sequence subfiles, sub_alignments, etc...). The OUT/
#			  directory stores all final results, adding in prefix 'align_' for file names. 
#			  The first step of the program saves the choices of the user, the second step split all
#			  sequence files in subfiles, each one is then aligned in one core using 'cmalign'. The
#			  third step is defined to merge all sub_alignments for each treated file in one global 
#			  alignment using 'cmalign' with the '--merge' option dedicated to this. Finally, to 
#			  obtain multiple alignments in FASTA format, we add a step to pass files from STOCKHOLM
#			  to FASTA format. And, at last, if the user want, we merge all alignments into one single
#			  file (stored in the OUT/ directory): Global_align.fasta.
#
#			  Details can be found in the source code below, based on all commentaries.
#			  1.5 Modifications: As we need to gain time for this step, we modify the step to merge 
#			  all sub_alignments for each treated file in one global alignment using 'cmalign' with 
#			  the '--merge' option. Previously, each file was merged with another named "X". Now, all
#			  merge steps are done in parallel using an idea given by Anne-Laure BLIEUX to gain time.
#			  Using a pyramidal system, all steps are now done faster than before. See details in the
#			  third step of the program with commentaries in 1.5 version
#
#			  2.0 Modifications: As we need to analyze also fungi data, we integrate also a new
#			  model to our program: the fungi (18S) model. This one was constructed using the defined
#			  command (see the model.cm to have the detail of the command line) with the seed of the
#			  RFAM alignment 01960 containing 79 sequences of eukarya (fungi and others). So, we give
#			  another choice of the user: to use this model during the alignment.
#
#			  2.1 Modifications: Little modification to fast another time the merging of alignments.
#			  To do this, we parallelize each merging step based on the 1.5 modifications. See details
#			  in the third step of the program with commentaries in 2.1 version
#
#			  2.5 Modis (12/04/12): This program was modified to be included in the workflow to fast the
#			  analysis. All modified steps are described in the program (e.g.: using the name of the
#			  input and a different location for the output file ans summary file, deletion of the 
#			  parallelization, as it was done by the main PERL program).
#
#			  2.6 Modifs (18/03/13): We modify and add a little element in the preparation of the 
#			  temporary sequence files, as if we create only one temporary file, we automatically 
#			  loose the first read analyzed. So, to avoid this problem, we add a little modification
#			  to correct this.
#
#			  2.7 Modifs (16/06/15): As we need to analyze also algae data targeting the 23S subunit, 
#			  we integrate also a new model to our program: the algea (23S) model. This one was 
#			  constructed using the defined command (see the model.cm to have the detail of the command 
#			  line) with the seed of the RFAM alignment RF02541 containing 102 sequences of bacteria 23S.
#			  So, we give another choice of the user: to use this model during the alignment.
#			  
#
#       AUTHOR:  Sebastien TERRAT <sebastien.terrat@dijon.inra.fr>
#      COMPANY:  UMR-MSE GenoSol platform, Team5
#      VERSION:  2.7
#     RELEASED:  04/08/2011
#     REVISION:  16/06/2015
#=======================================================================================================
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#=======================================================================================================

#Print errors and warnings during the implementation of the perl program
use strict;
use warnings;
#Needed to do some particular calculations like ceil() or floor()
use POSIX;
#Needed to manage the creation and the deletion of files and directories
use Inline;
use FindBin;

#Defined variables given as input by the program
my $model = lc($ARGV[0]);
my $cores = $ARGV[1] + 0;
my $file = $ARGV[2];
my $folder_input = $ARGV[3];
my $folder_output = $ARGV[4];
#Needed variables
my @seqs = ();
my @childs = ();
my @merge = ();
my @deleted_merge_files = ();
my @tmp = ();
my @file_name = ();
my $i = 0;
my $j = 0;
my $nb_seqs_tmp_files = 0;
my $pid = 0;
my $nb_tmp_files = 0;
my $line = "";
my $file_tmp = "";
my $command_align = "";
my $command_align_part = "";
my $command_align_final = "";
my $seq = "";
my $cmalign = "$FindBin::RealBin/../Data/Tools/INFERNAL/infernal-1.0.2/bin/cmalign";


#===================================================================================================
#2.5 : First step, using needed parameters (used matrix: bacteria or archaea), we define what model
#will be used for the alignment and verifying the number of cores used.
#===================================================================================================

#Verify the choice of the desired model given by the user and create the good $model data
#All models are stored in the Matrix directory and defined using parameters and sequences given 
#by INFERNAL developers.
	
#Choice of the bacteria model of covariance
if ($model eq "bacteria") { $model = "$FindBin::RealBin/../Prog_Perl/Matrix/bacteria_model.cm "; }
#Choice of the bacteria model of covariance
elsif ($model eq "fungi") { $model = "$FindBin::RealBin/../Prog_Perl/Matrix/fungi_model.cm "; }
#Choice of the archaea model of covariance
elsif ($model eq "archaea") { $model = "$FindBin::RealBin/../Prog_Perl/Matrix/archaea_model.cm "; }
#Choice of the archaea model of covariance
elsif ($model eq "algae") { $model = "$FindBin::RealBin/../Prog_Perl/Matrix/algae_model.cm "; }

#Verifying the 'int' type of the string given by the user
if ($cores =~ m/\d+/) {$cores = $cores + 0;}
else {$cores = 0;}
#Modifying the number of cores if needed
if (($cores < 1)||($cores > 32)){$cores = 16;}

#===================================================================================================
#Second step, creation of subfiles to parallelize the analysis based on the number of processors
#given by the user or defined by default. To do that, each file in the IN/ directory containing
#sequences in FASTA format have to be stored in tables to determine the number of sequences.
#===================================================================================================

#Needed empty variables
$seq = "";
@seqs = ();
@file_name = ();

#2.5: Open file needed to be treated
if ($folder_input eq "None")
{
	#2.5: Open file one given by the user and containing cleaned sequences
	open(F1, "Raw_data/$file")|| die "==> File Not Found: $file<==";
}
else
{
	#2.5: Open file one given by the user and containing cleaned sequences
	open(F1, "Result_files/$folder_input/$file")|| die "==> File Not Found: Result_files/$folder_input/$file<==";
}

#For every line of the file given by the user
LOOP01:while (<F1>)
{
	#If the line contains the character : ">"
	if ($_ =~ />/)
	{
		#Store it in $seq
		$seq = $_;
	}
	
	#If the line contains the sequence itself.
	else
	{
		#Store it in $seq with the description stored before
		$seq = $seq.$_;
		#Push it in the table @seqs
		push(@seqs, $seq);
		$seq = "";
	}
}
#Close the file used as input
close F1;


#Determine the number of sequences in each tmp file
$nb_seqs_tmp_files = floor($#seqs/$cores); 

#Loop to create each file with the adequate number of sequences
LOOP02:for($i=0;$i<$cores;$i++)
{
	#Create names of each subfiles stored in the Tmp/ directory
	$file_name[$i] = "Part.".$i.".".$file;
	#Open a new empty subfile in the Tmp/ directory
	open(F_tmp, ">Temporary_files/Part.$i.$file")|| die "==>Can't create tmp files";
	
	#2.6: correction of a small error due to the choice of one core, deleting one read (the first)
	#of the file and the array
	if ($cores == 1) { @tmp = @seqs; foreach $seq (@tmp) {printf(F_tmp $seq);} next LOOP02;}
	
	#Select a sub-array for each file based on several parameters:
	#The total number of seuences and the number of cores chosen by the user
	#If we create the last sequence subfile, we don't have to calculate the end of the
	#sub table @tmp
	if($i == ($cores-1)) {@tmp = @seqs[($i*$nb_seqs_tmp_files+1)..($#seqs)];}
	#If we create the first sequence subfile, we don't have to calculate the start of
	#the sub table @tmp
	elsif($i == 0) {@tmp = @seqs[0..(($i+1)*$nb_seqs_tmp_files)];}
	#For other sequence subfiles, we need to calculate each start and end of extraction
	else {@tmp = @seqs[($i*$nb_seqs_tmp_files+1)..(($i+1)*$nb_seqs_tmp_files)];}
	#Loop to print in each file the number of needed sequences in each subfile
	foreach $seq (@tmp) {printf(F_tmp $seq);}
	#Close the new created file
	close F_tmp;
}
#===================================================================================================
#Third step, using each temporary sequence file, the program will launch, based on user parameters, 
#the alignment program and produce temporary alignment files. These ones will be then merged
#using cmalign too (as it is already programmed by the developer of INFERNAL).
#===================================================================================================

#Needed command line with parameters to launch alignment steps (--hbanded: needed parameter
#based on developers information, --sub: to specify that with have partial sequences to
#align, --dna: to have the alignment is DNA format, not RNA, -1: to obtain alignments in 
#FASTA format, not in STOCKHOLM format, -q: quiet, to stop all printing data and details 
#for each aligned sequence, -o : to precise the filename where the alignment will be stored.
$command_align_part = "$cmalign --hbanded --sub --dna -1 -q -o ";		

#For each sequence subfile created before
LOOP03:foreach $file_tmp (@file_name)
{
	#Creation of a parent and a child process. One child for each found file needed treatment
	$pid = fork();
	#If the program is the parent! (only one)
	if ($pid) {push(@childs, $pid);}
	#If the program is a child (number as number of files)
	elsif ($pid == 0)
	{
		#2.5: We add the last parts of the command line (the name of the output file, the used 
		#model and the used sequence subfile.
		$command_align = $command_align_part."Temporary_files/align_".$file_tmp." ".$model." Temporary_files/".$file_tmp;
		#We finally execute this command line
		exec($command_align);
		exit(0);
	}
	#If we can't fork (create parent and child processes, die the process)
	else {die "couldn’t fork: $!\n";}
}
#The parent process wait for all child results
foreach (@childs) {waitpid($_, 0);}
#2.5: Loop to delete temporary files sequence files
foreach $file_tmp (@file_name){unlink("Temporary_files/$file_tmp");}



#1.5: Needed empty lists before merging of we have several files to treat
@deleted_merge_files = ();
@merge = ();
#1.5: Loop to create the table merge for merging files
foreach $file_tmp (@file_name)
{
	#1.5: we create the table @merge to store the filenames created in each step of merging
	push(@merge, "align_".$file_tmp);
}
#If the user have chosen more than one core
if($cores>1)
{
	#1.5: Needed empty variable to create temporary merged files
	$j = 0;
	#1.5: Loop to wait until only one file is kept
	while($#merge != 0)
	{
		#1.5: Needed variable to know the initial length of the table for this step
		$nb_tmp_files = $#merge;
		#1.5: Verifying if the table contain an even or an odd number of files
		if ((($#merge+1)%2)==0)
		{
			#1.5: Loop to merge files two by two
			for($i=0;$i<=$nb_tmp_files;$i=$i+2)
			{
				#2.5: Needed command line with parameters to launch merging of alignments
				#same parameters with one more: --merge to merge two alignments in one).
				#The global alignment will be stored in align_merge_filename (temporary file)
				$command_align_final = "$cmalign --hbanded --sub --dna -1 --merge -o Temporary_files/merge".$j."_".$file." ".$model." ";
				$command_align_final = $command_align_final."Temporary_files/$merge[$i]"." Temporary_files/$merge[$i+1]";
				#1.5: We add at the end of the table the new created file
				push (@merge, "merge".$j."_".$file);
				$j++;
				
				#Creation of a parent and a child process. One child for each found file needed treatment
				$pid = fork();
				#If the program is the parent! (only one)
				if ($pid) {push(@childs, $pid);}
				#If the program is a child (number as number of files)
				elsif ($pid == 0)
				{
					#1.5: We launch the merging of alignments
					exec($command_align_final);
					exit(0);
				}
				#If we can't fork (create parent and child processes, die the process)
		 		else {die "couldn’t fork: $!\n";}
			}
			#The parent process wait for all child results
			foreach (@childs) {waitpid($_, 0);}
		}
		#1.5: Verifying if the table contain an even or an odd number of files
		else
		{
			#1.5: Loop to merge files two by two, but don't treat the last file as we have an odd number of files
			for($i=0;$i<$nb_tmp_files;$i=$i+2)
			{
				#2.5: Needed command line with parameters to launch merging of alignments
				#same parameters with one more: --merge to merge two alignments in one).
				#The global alignment will be stored in align_merge_filename (temporary file)
				$command_align_final = "$cmalign --hbanded --sub --dna -1 --merge -o Temporary_files/merge".$j."_".$file." ".$model." ";
				$command_align_final = $command_align_final."Temporary_files/$merge[$i]"." Temporary_files/$merge[$i+1]";
				#1.5: We add at the end of the table the new created file
				push (@merge, "merge".$j."_".$file);
				$j++;
				
				#Creation of a parent and a child process. One child for each found file needed treatment
				$pid = fork();
				#If the program is the parent! (only one)
				if ($pid) {push(@childs, $pid);}
				#If the program is a child (number as number of files)
				elsif ($pid == 0)
				{
					#1.5: We launch the merging of alignments
					exec($command_align_final);
					exit(0);
				}
				#If we can't fork (create parent and child processes, die the process)
		 		else {die "couldn’t fork: $!\n";}
			}
			#The parent process wait for all child results
			foreach (@childs) {waitpid($_, 0);}
		}
		#1.5: We add also the names of files that we need to delete before the next step
		@deleted_merge_files = (@deleted_merge_files, @merge[0..$i-1]);
		#1.5: We kept only the files needed based on $i value
		@merge = @merge[$i..$#merge];
	}
	#2.5: Copy the global alignment in align_merge_filename to merge it another time, etc...
	system("cp Temporary_files/".$merge[$#merge]." Temporary_files/align_merge_".$file);
	#2.5: Delete the file containing the global alignment with a tmp name
	unlink("Temporary_files/$merge[$#merge]");
	#2.5: Deletion of all sub_alignments in the Tmp/directory
	foreach $file_tmp (@deleted_merge_files) {unlink("Temporary_files/$file_tmp");}
}
#If we used only one core, the global alignment is already done
else
{
	#2.5: Copy the align_X_filename in align_merge_filename
	system("cp Temporary_files/align_".$file_name[0]." Temporary_files/align_merge_".$file);
	#2.5: Deletion of all sub_alignments in the Tmp/directory
	foreach $file_tmp (@file_name) {unlink("Temporary_files/align_$file_tmp");} 
}


#===================================================================================================
#Fourth step, modification of each global alignment to obtain alignments in FASTA format, not in
#STOCKHOLM format for further analysis.
#===================================================================================================

#2.5: Opening two files, the global alignment align_merge_filename and the final result file
#for the user
open(F_merge, "Temporary_files/align_merge_$file")|| die "==> Can't open the merge file";
open(F_result, ">Result_files/$folder_output/Align_$file")|| die "==> Can't create final alignment file";

#For each line of the file
BOUCLE5:while(<F_merge>)
{
	#If we found the # character, it's a descriptive line of the STOCKHOLM format
	if ($_ =~ m/#/)
	{
		#If we found the consensus sequence or the profile sequence, we keep them
		if (($_ =~ s/GC RF/GC_RF/)||($_ =~ s/GC SS/GC_SS/))
		{
			#Modifying the line to obtain a FASTA format of the alignment
			$line = $_;
			$line =~ s/\s+/\n/;
			#Storying it the final result file
			printf(F_result ">$line");
		}
	}
	#If we found only a \n or a line with '//', we don't keep them
	elsif (($_ eq "\n")||($_ =~ /\/\//)) {next BOUCLE5;}
	#Or, for alignment sequence lines, we modify them to obtain a FASTA format.
	else
	{
		$line = $_;
		$line =~ s/\s+/\n/;
		printf(F_result ">$line");
	}
}
#Close the two files
close F_merge;
close F_result;

#2.5: Deleting the align_merge file stored in the Tmp directory
unlink("Temporary_files/align_merge_$file");

#2.5: We pipe the name of the created file (the result file) for the main program
print "Align_$file";

