#=======================================================================================================
#         FILE:  Phylo_tree_computation_1.0.pl
#        USAGE:  perl Phylo_tree_computation_1.0.pl arguments
#	ARGUMENTS:  - Taxonomy input folder  [FOLDER_NAME],
#			  - Alignment input folder [FOLDER_NAME],
#			  - The output folder where trees will be stored [FOLDER_NAME],
#			  - Chosen threshold for the taxonomic step (confidence or similarity) [0-100],
#			  - Chosen taxonomic level [phylum, class, order, family, genus or all],
#			  - The chosen model of covariance used during the alignment step [bacteria, archaea or 
#			  fungi].
#  DESCRIPTION:  This PERL program was developed to define easily and efficiently phylogenetic trees 
#			  at all defined taxonomic levels (phylum, class, order, family, genus). To do this, we
#			  first read all taxonomic files, and we catch all names of groups found at one taxonomic
#			  level (we keep also the best read for each group among all reads and files). Using 
#			  these read IDs, we search and store the aligned reads and create a global alignment of
#			  all kept groups. Finally, using this formatted global alignment, we create a 
#			  phylogenetic tree with the FastTree program.
#			  The result file (the phylogenetic tree) will be stored in a dedicated folder given by 
#			  the main program.
#			  (see details in the program).
#
# REQUIREMENTS:  Perl language only
#       AUTHOR:  Sebastien TERRAT <sebastien.terrat@dijon.inra.fr>
#      COMPANY:  GenoSol platform (UMR 1347 INRA/uB/AgroSup)
#      VERSION:  1.0
#    REALEASED:  XX
#     REVISION:  XX
#=======================================================================================================
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#=======================================================================================================

#!/usr/bin/perl

#Print errors and warnings during the implementation of the perl program
use strict;
use warnings;
use FindBin;

#Defined needed variables
my $folder_taxonomy = "TAXONOMY";
my $folder_alignment = "CLEAN_INFERNAL_ALIGNMENT";
my $folder_output = "Summary_files/Phylogenetic_trees";
my $taxo_cutoff = 80/100;
my $level = "genus";
my $model = "fungi";

#Defined variables
my $max_length = 0;
my $nb_tmp_files = 0;
my ($i,$j) = (0,0);
my $kept_model = $model;
my $file = "";
my $align_seq = "";
my $taxo_name = "";
my $align_file = "";
my $command_align_final = "";
my $line = "";
my $FastTree_result = "";
my $chosen_names = "";
my %taxo_selected_reads = ();
my %kept_aligned_files = ();
my @tmp = ();
my @tot_aligned_files = ();
my @deleted_merge_files = ();
my @kept_aligned_files = ();

#Here, we define the absolute path to find the FastTree program stored efficiently to launch 
#needed analyses.
my $folder_Tools = "$FindBin::RealBin/../Data/Tools";

if(defined($ARGV[0]))
{
	open (F1,"$ARGV[0]") || die "==> File Not Found: $ARGV[0]<==";
	while(<F1>) {$chosen_names .= $_;}
	close (F1);
}
else { }
$chosen_names =~ s/\./\-/;

#===================================================================================================
# First step, using given parameters (used matrix: bacteria, archaea or fungi), we define what model
# will be used for the alignment and verifying the number of cores used. Also, using the
# $folder_alignment information given by the main program, we will catch and store all alignment
# files names to use them in another step of analysis.
#===================================================================================================

#Verify the choice of the desired model given by the user and create the good $model data
#All models are stored in the Matrix directory and defined using parameters and sequences given 
#by INFERNAL developers.
	
#Choice of the bacteria model of covariance
if ($model eq "bacteria") { $model = "Prog_Perl/Matrix/bacteria_model.cm "; }
#Choice of the bacteria model of covariance
elsif ($model eq "fungi") { $model = "Prog_Perl/Matrix/fungi_model.cm "; }
#Choice of the archaea model of covariance
elsif ($model eq "archaea") { $model = "Prog_Perl/Matrix/archaea_model.cm "; }

#We open the defined directory to find the files with alignment results
opendir (DIR, "Result_files/$folder_alignment/") || die "can't opendir Result_files/$folder_alignment/: $!";
#We will then read and store all file names to use them in another step
LOOP01:while (defined($file = readdir(DIR)))
{
	if ($file =~ /^\.\.?$/) { next LOOP01; }
	else { push(@tot_aligned_files, $file); }
}

#=======================================================================================================
# Second step, based on the choice of the user (genus, class, phylum or all), we will read all result 
# files to find and define which groups will be displayed in the phylogenetic tree(s), using one read 
# for each group).
#=======================================================================================================

#We open the defined directory to find the files with taxonomic results
opendir (DIR, "Result_files/$folder_taxonomy/") || die "can't opendir Result_files/$folder_taxonomy/: $!";
#We will then read all result files and find the perfect affiliated sequence among all sequences to 
#compute the phylogenetic tree. 
LOOP02:while (defined($file = readdir(DIR)))
{
	#If the file name is "." or "..", we go for the next file
	if ($file =~ /^\.\.?$/) { next LOOP02; }
	else
	{
		#We then open the foudn file in the $folder_taxonomy
		open (F1,"Result_files/$folder_taxonomy/$file") || die "==> File Not Found: Result_files/$folder_taxonomy/$file<==";
		#For each line found in the opened file
		LOOP03:while(<F1>)
		{
			#We first treat some elements in the line to easily read them
			$_ =~ s/"//g;
			$_ =~ s/\t-\t/\t\t/g;
			#We then search for the sequence ID, it's score and it's associated name for the defined level
			#We avoid using an "if" the lines without defined names at the chosen level (phylum, class, etc...).
			if ($_ =~ m/([A-Za-z0-9\_\=]*)\t\t.*\t([A-Za-z0-9\_\-\/]*)\t$level\t(\d\.\d*)\t/)
			{
				#We put this test as for sequences with Unknown at 0.00, it can be defined as Unclassified for Bacteria
				if (($2 eq "Unknown")&&($3 == 0.00)&&(uc($kept_model) eq "BACTERIA"))
				{
					$taxo_selected_reads{"Unclassified"} = "1.00"."###".$1."###".$file;
				}
				elsif (($2 eq "Unknown")&&(uc($kept_model) eq "FUNGI"))
				{
					$taxo_selected_reads{"Unknown"} = "1.00"."###".$1."###".$file;
				}
				#If the name is the first found
				elsif (not exists($taxo_selected_reads{$2}))
				{
					#We automatically add it to the hash table %taxo_selected_reads
					$taxo_selected_reads{$2} = $3."###".$1."###".$file;
				}
				#If it is not the first time we found it, we will test it
				else
				{
					#Indeed, we split the already information stored using the @tmp array
					@tmp = split("###", $taxo_selected_reads{$2});
					#And, if the score for the new read is higher than the stored one, we will
					#keep the highest one
					if ($tmp[0] <= $3)
					{
						$taxo_selected_reads{$2} = $3."###".$1."###".$file;
					}
				}
			}
			else { }
		}
		#We then close the read file
		close (F1);
	}
}
#We finally close the $taxonomy_folder directory
closedir (DIR);




if(defined($ARGV[0]))
{
	foreach $taxo_name (keys(%taxo_selected_reads))
	{
		if($chosen_names =~ m/$taxo_name/)
		{
			next;
		}
		else
		{
			delete( $taxo_selected_reads{$taxo_name} );
		}
	}
}
else
{
}

#=======================================================================================================
# Third step, we will find the needed reads of all chosen files that will be used by the FastTree
# program to develop the phylogenetic tree. But, first of all, we will store them all in several 
# dedicated temporary files to use use and transform them in STOCKHOLM format to finally merge them.
#=======================================================================================================

#For each taxonomic name found in the list of all reads
LOOP04:foreach $taxo_name (keys(%taxo_selected_reads))
{
	#We first need to compare the length of all taxonomic names and keep the longest
	if(length($taxo_name) > $max_length)
	{
		$max_length = length($taxo_name);
	}
	else {}
	#We split in @tmp the information stored in %taxo_selected_reads for the chosen name
	@tmp = split("###", $taxo_selected_reads{$taxo_name});
	#We verify that the threshold is lower than the score given for the read (the highest one found)
	if ($tmp[0] > $taxo_cutoff)
	{
		#We catch the file name from the stored information
		$file = $tmp[2];
		#We treat the file name to simplify it
		$file =~ s/[LMH]preprocess_//g;	$file =~ s/Rdm_//g;		$file =~ s/Taxo_//g;
		$file =~ s/Align_//g;			$file =~ s/Clust_//g;	$file =~ s/Derep_//g;
		$file =~ s/Clean_//g;			$file =~ s/Hunt_//g;	$file =~ s/SILVA_//g;
		$file =~ s/RDP_//g;
		#Then, among all aligned files found in the @tot_aligned_files
		LOOP05:foreach $align_file (@tot_aligned_files)
		{
			#If we found the needed file
			if ($align_file =~ m/$file/)
			{
				#We keep it in the hash table %kept_aligned_files
				$kept_aligned_files{$align_file}++;
				#We open this file (read) and a new temporary file (append)
				open (F1,"Result_files/$folder_alignment/$align_file") || die "==> File Not Found: Result_files/$folder_alignment/$align_file <==";
				open (F_TMP,">>Temporary_files/TMP_Phylo_$align_file") || die "==> File Not Found: Temporary_files/TMP_Phylo_$align_file <==";
				#We use this loop to read all lines and find the searched one corresponding
				#to the needed read
				LOOP06:while(<F1>)
				{
					#If the line contains the ID of the read
					if($_ =~ m/^>$tmp[1]/)
					{
						#We catch the next line (the aligned read)
						$align_seq = <F1>;
						#And we print the taxonomic name and the aligned read in the temporary file
						printf(F_TMP ">$taxo_name\n$align_seq");
					}
				}
				#We close the two files (the input file and the temporary file)
				close (F_TMP);
				close (F1);
			}
		}
	}
	#If the read is lower than the defined threshold, we go for the next taxonomic name
	else { next LOOP04; }
}

#We catch the names of the kept aligned files to treat them only in further steps and empty the hash 
#table.
@kept_aligned_files = keys(%kept_aligned_files);

#=======================================================================================================
# Fourth step, we treat all temporary aligned files to format them in the STOCKHOLM format efficiently
# and easily.
#=======================================================================================================

#For each file that needs a formatting
LOOP07:foreach $align_file (@kept_aligned_files)
{
	#We open the new temporary file and first print several descriptive lines
	open (F_OUT,">Temporary_files/Align_Phylo_$align_file") || die "==> File Not Found: Temporary_files/Align_Phylo_$align_file <==";
	printf(F_OUT "# STOCKHOLM 1.0\n#=GF AU Infernal 1.0.2\n\n");
	#We then open the previous temporary file
	open (F_TMP,"Temporary_files/TMP_Phylo_$align_file") || die "==> File Not Found: Temporary_files/TMP_Phylo_$align_file <==";
	#For each line found in this temporary file
	LOOP08:while(<F_TMP>)
	{
		# If it's a descriptive line
		if($_ =~ m/^>/)
		{
			#We first delete some not-needed characters
			$_ =~ s/^>//; $_ =~ s/\n//;
			#Based on the max_length defined previously, we add spaces at the end of the descriptor
			while (length($_) < $max_length + 1) { $_ = $_." ";}
			#We then print the descriptive line
			printf(F_OUT $_);
		}
		#If it's an aligned sequence, we directly print it 
		else { printf(F_OUT $_); }
	}
	#We finally close the temporary file
	close (F_TMP);
	#We then open the initial alignment file to catch the two needed lines (the consensus and the structure of the consensus read)
	open (F1,"Result_files/$folder_alignment/$align_file") || die "==> File Not Found: Result_files/$folder_alignment/$align_file <==";
	#To do this, we read all the file
	LOOP09:while(<F1>)
	{
		#If we found these two particular lines
		if(($_ =~ m/#=GC_SS_cons/)||($_ =~ m/#=GC_RF/))
		{
			#We first delete or replace some particular characters
			$_ =~ s/^>//; $_ =~ s/\n//; $_ =~ s/GC_/GC /;
			#e add spaces to them if needed
			while (length($_) < $max_length + 1) { $_ = $_." ";}
			#We then print the descriptive line and also the aligned structure and
			#consensus sequence
			printf(F_OUT $_);
			$_ = <F1>;
			printf (F_OUT $_);
		}
	}
	#We finally close the initial alignment file
	close (F1);
	#We add the needed characters to specify the end of the alignment
	printf(F_OUT "//");
	#And we close the temporary alignment file
	close (F_OUT);
}

#Last loop of this part to delete
LOOP10:foreach $align_file (@kept_aligned_files) { unlink("Temporary_files/TMP_Phylo_$align_file"); }

#===================================================================================================
# Fifth step, using each temporary sequence file, the program will launch, based on user parameters, 
# the alignment program and produce temporary alignment files. These ones will be then merged
# using cmalign too (as it is already programmed by the developer of INFERNAL).
#===================================================================================================

#Needed empty variable to create temporary merged files
$j = 0;
$file = "Align_Tmp_file.fasta";
#Loop to wait until only one file is kept
LOOP11:while($#kept_aligned_files != 0)
{
	#Needed variable to know the initial length of the table for this step
	$nb_tmp_files = $#kept_aligned_files;
	#Verifying if the table contain an even or an odd number of files
	if ((($#kept_aligned_files+1)%2)==0)
	{
		#Loop to merge files two by two
		LOOP12:for($i=0;$i<=$nb_tmp_files;$i=$i+2)
		{
			#Needed command line with parameters to launch alignment steps (--hbanded: needed parameter
			#based on developers information, --sub: to specify that with have partial sequences to
			#align, --dna: to have the alignment is DNA format, not RNA, -1: to obtain alignments in 
			#FASTA format, not in STOCKHOLM format, -q: quiet, to stop all printing data and details 
			#for each aligned sequence, -o : to precise the filename where the alignment will be stored.
			#One more parameter ot merge alignments: --merge to merge two alignments in one.
			#The global alignment will be stored in align_merge_filename (temporary file)
			$command_align_final = "cmalign --hbanded --sub --dna -1 --merge -o Temporary_files/Align_Phylo_merge".$j."_".$file." ".$model." ";
			$command_align_final = $command_align_final."Temporary_files/Align_Phylo_$kept_aligned_files[$i]"." Temporary_files/Align_Phylo_$kept_aligned_files[$i+1]";
			#We add at the end of the table the new created file
			push (@kept_aligned_files, "merge".$j."_".$file);
			$j++;
			#We launch the merging of alignments
			system($command_align_final);
		}
	}
	#Verifying if the table contain an even or an odd number of files
	else
	{
		#Loop to merge files two by two, but don't treat the last file as we have an odd number of files
		LOOP13:for($i=0;$i<$nb_tmp_files;$i=$i+2)
		{
			#Needed command line with parameters to launch merging of alignments
			#same parameters with one more: --merge to merge two alignments in one).
			#The global alignment will be stored in align_merge_filename (temporary file)
			$command_align_final = "cmalign --hbanded --sub --dna -1 --merge -o Temporary_files/Align_Phylo_merge".$j."_".$file." ".$model." ";
			$command_align_final = $command_align_final."Temporary_files/Align_Phylo_$kept_aligned_files[$i]"." Temporary_files/Align_Phylo_$kept_aligned_files[$i+1]";
			#We add at the end of the table the new created file
			push (@kept_aligned_files, "merge".$j."_".$file);
			$j++;
			#We launch the merging of alignments
			system($command_align_final);
		}
	}
	#We add also the names of files that we need to delete before the next step
	@deleted_merge_files = (@deleted_merge_files, @kept_aligned_files[0..$i-1]);
	#We kept only the files needed based on $i value
	@kept_aligned_files = @kept_aligned_files[$i..$#kept_aligned_files];
}
#Copy the global alignment in align_merge_filename to merge it another time, etc...
system("cp Temporary_files/Align_Phylo_".$kept_aligned_files[$#kept_aligned_files]." Temporary_files/Complete_TMP_".$file);
#Delete the file containing the global alignment with a tmp name
unlink("Temporary_files/Align_Phylo_$kept_aligned_files[$#kept_aligned_files]");
#Deletion of all sub_alignments in the Temporary directory
LOOP14:foreach $file (@deleted_merge_files) { unlink("Temporary_files/Align_Phylo_$file"); }

#===================================================================================================
# Last step, modification of the global alignment to obtain an alignment in FASTA format, not in
# STOCKHOLM format, needed for the creation of the phylogenetic tree.
#===================================================================================================

#Opening two files, the global alignment in STOCKHOLM format and the new alignment file in FASTA format
open(F_merge, "Temporary_files/Complete_TMP_".$file)|| die "==> Can't open the merge file";
open(F_result, ">Temporary_files/Complete_Phylo_$file")|| die "==> Can't create final alignment file";

#For each line of the alignment file in STOCKHOLM format
LOOP15:while(<F_merge>)
{
	#If we found the # character, it's a descriptive line of the STOCKHOLM format, we avoid it
	if ($_ =~ m/#/) { next LOOP15; }
	#If we found only a \n or a line with '//', we don't keep them
	elsif (($_ eq "\n")||($_ =~ /\/\//)) { next LOOP15; }
	#Or, for alignment sequence lines, we modify them to obtain a FASTA format.
	else
	{
		$line = $_;
		$line =~ s/\s+/\n/;
		$line =~ s/\./\-/g;
		printf(F_result ">$line");
	}
}
#Close the two files
close F_merge;
close F_result;

#Deleting the align_merge file stored in the Temprary directory
unlink("Temporary_files/Complete_TMP_$file");

#Then, we launch the FastTree program (stored in the Prog_Perl folder), with these options:
#	-nt (indicate that we gvie as input a multiple nucleotide alignment,
#	-spr 4 (to define the number of rounds of SPR (higher phylogenetic tree quality),
#	-mlacc 2 (mlacc 2 or -mlacc 3 to always optimize all 5 branches at each NNI, and to optimize all 5 branches in 2 or 3 rounds), higher phylogenetic tree quality,
#	-slownni (to turn off heuristics to avoid constant subtrees (affects both ML and ME NNIs)), higher phylogenetic tree quality,
#	-quiet (do not write to standard error during normal operation (no progress indicator, no options summary, no likelihood values, etc.).
$level = "Phylogenetic_tree_".$level.".tre";
$FastTree_result = `$folder_Tools/FastTree/FastTree -nt -spr 4 -mlacc 2 -slownni -quiet Temporary_files/Complete_Phylo_$file > Summary_files/Phylogenetic_trees/$level`;

#Deleting the align_merge file stored in the Temprary directory
unlink("Temporary_files/Complete_Phylo_$file");
