#=======================================================================================================
#         FILE:  PhyloTree_computation_known_1.0.pl
#        USAGE:  perl PhyloTree_computation_known_1.0.pl [ARGUMENTS]
#	ARGUMENTS:  - $ARGV[0]: Model (bacteria, archaea, or fungi) used to realize the global alignment
#			  - $ARGV[1]: Level (phylum, class, order, family, genus, etc...) to analyze the good
#			    taxonomic file.
#			  - $ARGV[2]: Database (RDP or Silva), to know where to search the known sequences.
#  DESCRIPTION:  This program first search in the Summary_files folder to find Taxonomic summary files,
#			  then, it store only taxonomic groups with at least one sequence in all samples. Using
#			  these taxonomic names, it select randomly one sequence in the database of known 
#			  sequences, to represent each taxononic group. It laucnhed then a global alignment with
#			  Infernal, format it, and launch the creation of a FastTree phylogenetic tree using the 
#			  global alignment.
#
# REQUIREMENTS:  Perl language only
#       AUTHOR:  Sebastien TERRAT <sebastien.terrat@dijon.inra.fr>
#      COMPANY:  GenoSol platform (UMR 1347 INRA/uB/AgroSup)
#      VERSION:  1.0
#    REALEASED:  30/01/2014
#     REVISION:  XX
#=======================================================================================================
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#=======================================================================================================

#!/usr/bin/perl

#Print errors and warnings during the implementation of the perl program
use strict;
use warnings;
use FindBin;
#Defined needed variables by the user as input arguments
my $model = $ARGV[0];
my $level = $ARGV[1];
my $database = $ARGV[2];

#Here, we define the absolute path to find the taxonomic databases for all taxonomic steps
#(Bacteria with RDP or SILVA, and Fungi with SILVA).
my $folder_struct = "$FindBin::RealBin/../Data/Databases/".uc($model)."/".uc($database)."/";
my $folder_Tools = "$FindBin::RealBin/../Data/Tools";

#Needed command line with parameters to launch alignment steps (--hbanded: needed parameter
#based on developers information, --sub: to specify that with have partial sequences to
#align, --dna: to have the alignment is DNA format, not RNA, -1: to obtain alignments in 
#FASTA format, not in STOCKHOLM format, -q: quiet, to stop all printing data and details 
#for each aligned sequence, -o : to precise the filename where the alignment will be stored.
my $command_align_part = "cmalign --hbanded --sub --dna -1 -q -o ";


#Needed empty variables
my $i = 0;
my $verif = 0;
my $done = 0;
my $file = "";
my $line = "";
my $name = "";
my $command_align = "";
my @taxo_files = ();
my @taxo_names = ();
my @tmp = ();
my %chosen_names = ();


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Function defined to search, using three arguments (the name of the taxonomic group, the level and the
# database) the good sequence to represent in the phylogenetic tree the taxonomic group. 
sub SearchDB
{
	#Needed variables (input or just defined).
	my $name = $_[0];
	my $level = $_[1];
	my $database = $_[2];
	my $verif = 0;
	my $seq = "NONE";
	
	if (($name eq "Alphaproteobacteria")||($name eq "Betaproteobacteria")||($name eq "Deltaproteobacteria")
	||($name eq "Gammaproteobacteria")||($name eq "Epsilonproteobacteria"))
	{
		$level = "class";
	}
	
	#We open the file of known and cleaned sequences
	open(F_DB, $folder_struct."Raw_data_FASTA/".$database) || die "can't open the database file<==";
	open(F_OUTX,">>Temporary_files/Extracted_reads.txt") || die "==> File Not Found: Temporary_files/Extracted_reads.txt <==";
	#We search in it
	LOOP06:while(<F_DB>)
	{
		#If the line is a descriptive one
		if($_ =~ m/^>/)
		{
			#We avoid the treatment of the file if we already found a representative sequence
			if($verif == 1) { close(F_DB); return $seq; }
			#If not, we search for a representative sequence
			if($_ =~ m/\;$name\($level\)\;/)
			{
				printf(F_OUTX $_);
				$seq = ">$name\n";
				$verif = 1;
			}
		}
		#We store the sequence itself
		else { if($verif == 1) { $_ =~ s/U/T/g; printf(F_OUTX $_); $seq .= $_; } }
	}
	#We cloe and return the sequence and its descriptive line.
	close(F_DB);
	close(F_OUTX);
	return $seq;
}
# End of the function
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#===================================================================================================
# First step, using given parameters (used matrix: bacteria, archaea or fungi), we define what model
# will be used for the alignment.
#===================================================================================================

#Verify the choice of the desired model given by the user and create the good $model data
#All models are stored in the Matrix directory and defined using parameters and sequences given 
#by INFERNAL developers.
	
#Choice of the bacteria model of covariance
if ($model eq "bacteria") { $model = "Prog_Perl/Matrix/bacteria_model.cm "; }
#Choice of the bacteria model of covariance
elsif ($model eq "fungi") { $model = "Prog_Perl/Matrix/fungi_model.cm "; }
#Choice of the archaea model of covariance
elsif ($model eq "archaea") { $model = "Prog_Perl/Matrix/archaea_model.cm "; }

#===================================================================================================
# Second step, we search for taxonomic names with at least one read in one sample.
#===================================================================================================

#We open the summary_files directory to find the files with taxonomy results
opendir (DIR, "Summary_files/") || die "can't opendir Summary_files/: $!";
#We will then read and store all file names to use them in another step
LOOP01:while (defined($file = readdir(DIR)))
{
	if ($file =~ m/^\.\.?$/) { next LOOP01; }
	else { if ($file =~ m/Taxonomy_results/) { push(@taxo_files, $file); } }
}

#We then search for the taxonomic file corresponding to the given level
LOOP02:foreach $file (@taxo_files)
{
	if ($file =~ m/_$level/)
	{
		#We open the chosen file
		open (F_TAXO,"Summary_files/$file") || die "==> File Not Found: Summary_files/$file <==";
	}
}
#We also test for the existence of the file with the good level, or we stop the program
if(not fileno(F_TAXO)) { die "The indicated level of taxonomy was unavailable, please retry with another level\n";}

#For each line of the taxonomic file
LOOP03:while(<F_TAXO>)
{
	#We delete the last character (\n)
	chomp($_);
	#If the line is the line with wanted taxonomic names
	if($_ =~ m/^Taxonomy_$level/)
	{
		#We store the names in the array @taxo_names
		@taxo_names = split("\t", $_);
		#We then read all the file for result for each sample
		LOOP04:while(<F_TAXO>)
		{
			#We delete the last character (\n)
			chomp($_);
			#We split the line based on tabulations
			@tmp = split("\t", $_);
			#Loop to search for taxonomic names with at least one read found, we store them in
			# the hash table %chosen_names
			LOOP05:for($i=1;$i<=$#tmp;$i++)
			{
				if($tmp[$i] > 0)
				{
					if ($taxo_names[$i] eq "Proteobacteria")
					{
						$chosen_names{"Alphaproteobacteria"}++;
						$chosen_names{"Betaproteobacteria"}++;
						$chosen_names{"Deltaproteobacteria"}++;
						$chosen_names{"Gammaproteobacteria"}++;
						$chosen_names{"Epsilonproteobacteria"}++;
					}
					else
					{
						$chosen_names{$taxo_names[$i]}++;
					}
				}
			}
		}
	}
}
close(F_TAXO);
#We finally store needed names in the good array @taxo_names
@taxo_names = keys(%chosen_names);

#===================================================================================================
# Third step, after the selection of all known names, we searched in the databases the needed 
# sequences.
#===================================================================================================

#File to store sequences in FASTA format
open(F_OUT,">Temporary_files/Selected_reads.txt") || die "==> File Not Found: Temporary_files/Test.txt <==";
#For each name
LOOP05:foreach $name (@taxo_names)
{
	#We avoid those not needed
	if(($name eq "Unclassified")||($name eq "Unknown")||($name eq "Environmental")||($name eq "Undefined"))
	{
		next LOOP05;
	}
	#We use the SearchDB subfunction to search efficiently the sequences, first in the C database, and then
	#in the EI database.
	$line = &SearchDB($name, $level, "C_SILVA_DB_R114.fasta");
	if ($line ne "NONE") { printf(F_OUT $line); next LOOP05; }
	$line = &SearchDB($name, $level, "EI_SILVA_DB_R114.fasta");
	if ($line ne "NONE") { printf(F_OUT $line); next LOOP05; }
}
close(F_OUT);

#We add the last parts of the command line (the name of the output file, the used model and the used sequence subfile.
$command_align = $command_align_part."Temporary_files/Align_selected_reads.txt ".$model." Temporary_files/Selected_reads.txt";
#We finally execute this command line
system($command_align);

#===================================================================================================
# Last step, modification of the global alignment to obtain an alignment in FASTA format, not in
# STOCKHOLM format, needed for the creation of the phylogenetic tree.
#===================================================================================================

#Opening two files, the global alignment in STOCKHOLM format and the new alignment file in FASTA format
open(F_merge, "Temporary_files/Align_selected_reads.txt")|| die "==> Can't open the alignment file";
open(F_result, ">Temporary_files/Clean_Align_selected_reads.txt")|| die "==> Can't create final alignment file";

#For each line of the alignment file in STOCKHOLM format
LOOP06:while(<F_merge>)
{
	#If we found the # character, it's a descriptive line of the STOCKHOLM format, we avoid it
	if ($_ =~ m/#/) { next LOOP06; }
	#If we found only a \n or a line with '//', we don't keep them
	elsif (($_ eq "\n")||($_ =~ /\/\//)) { next LOOP06; }
	#Or, for alignment sequence lines, we modify them to obtain a FASTA format.
	else
	{
		$line = $_;
		$line =~ s/\s+/\n/;
		$line =~ s/\./\-/g;
		printf(F_result ">$line");
	}
}
#Close the two files
close F_merge;
close F_result;

#Then, we launch the FastTree program (stored in the Prog_Perl folder), with these options:
#	-nt (indicate that we gvie as input a multiple nucleotide alignment,
#	-spr 4 (to define the number of rounds of SPR (higher phylogenetic tree quality),
#	-mlacc 2 (mlacc 2 or -mlacc 3 to always optimize all 5 branches at each NNI, and to optimize all 5 branches in 2 or 3 rounds), higher phylogenetic tree quality,
#	-slownni (to turn off heuristics to avoid constant subtrees (affects both ML and ME NNIs)), higher phylogenetic tree quality,
#	-quiet (do not write to standard error during normal operation (no progress indicator, no options summary, no likelihood values, etc.).
$line = `$folder_Tools/FastTree/FastTree -nt -spr 4 -mlacc 2 -slownni -quiet Temporary_files/Clean_Align_selected_reads.txt > Summary_files/Phylogenetic_Tree_$level.tre`;
#We delete the not needed files
unlink("Temporary_files/Selected_reads.txt");
unlink("Temporary_files/Align_selected_reads.txt");
unlink("Temporary_files/Clean_Align_selected_reads.txt");
exit(0);
